import gzip
import copy
import pickle
import os.path
import sys

import numpy as np

from tqdm import tqdm
import torch
from torch import optim
from torch.utils.tensorboard import SummaryWriter
from torchvision import datasets, transforms
from torch.utils.data import DataLoader
from torchvision import datasets, transforms
import torch.nn as nn
import torch.nn.functional as F

from options import args_parser
from model import vae_mini, vie

from torchvision.models import resnet18

from utils.cifar10_dataset import CIFAR10SubSet
from utils.cifar100_dataset import CIFAR100SubSet

""" TESTING on CIFAR-10"""

def get_schedulers(scheduler, optimizer, milestones=[30,80], gamma=0.5, T_max=10, lr_mul=0.001, d_model=10, n_warmup_steps=5):
    if scheduler == "step":
        return torch.optim.lr_scheduler.StepLR(optimizer, 30, gamma=gamma)
    elif scheduler == "cosine":
        return torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max)
    elif scheduler == "exponential":
        return torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=gamma)


def main():
    """Test SplitNN"""
    args = args_parser()
    lr = 1e-2
    epoch_max = 500
    epoch_finetune = 5
    bs = 32
    Q = 1
    if args.dataset == 'cifar100':
        num_class = 100
    else:
        num_class = 10

    train_size = 50000

    criterion = torch.nn.CrossEntropyLoss()
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    print("Training on device", device)


    net = resnet18()
    net.fc = nn.Linear(net.fc.in_features, num_class)

    # decoder = nn.Sequential(
    #         # input is Z
    #         nn.ConvTranspose2d(64, 3, 8, 2, 3),
    #         nn.Sigmoid())

    class Flatten(nn.Module):
        def __init__(self):
            super(Flatten, self).__init__()
            
        def forward(self, x):
            x = x.view(x.size(0), -1)
            return x

    modules = list(net.children())[:args.num_head_layer]
    head = nn.Sequential(*modules)

    vae = vie(feature_volume = 64*8*8, z_size = 64*8*8)


    modules = list(net.children())[:-4]
    encoder = nn.Sequential(*modules)
    modules = list(encoder.children())[args.num_head_layer:]
    encoder = nn.Sequential(*modules)

    modules = list(net.children())[-4:-1]
    clf = nn.Sequential(*[*modules, Flatten(), list(net.children())[-1]])

    encoder.to(device)
    clf.to(device)
    head.to(device)
    vae.to(device)



    '''
    if save_models_filename:
        if os.path.exists(save_models_filename):
            print(f"Restoring models from {save_models_filename}")
            data = torch.load(save_models_filename)
            # print("net1", net1.state_dict().keys())
            # print("data['net1']", data["net1"].keys())
            # net1.load_state_dict(data["net1"])
            # net2.load_state_dict(data["net2"])
            net1_a = data["net1_a"]
            net1_b = data["net1_b"]
            net2 = data["net2"]
    '''
    optim_head = optim.SGD(head.parameters(), lr=lr, momentum=0.9)
    optim_vae = optim.SGD(vae.parameters(), lr=lr, momentum=0.9)
    optim_encoder = optim.SGD(encoder.parameters(), lr=lr, momentum=0.9)
    optim_clf = optim.SGD(clf.parameters(), lr=lr, momentum=0.9)


    '''
    transform_train = transforms.Compose(
        [
            transforms.ToTensor(),
            #transforms.ToPILImage(),
            #transforms.Pad(4, padding_mode="reflect"),
            #transforms.RandomCrop(32),
            #transforms.RandomHorizontalFlip(),
            #transforms.ToTensor(),
            transforms.Normalize(
                mean=[x / 255.0 for x in [125.3, 123.0, 113.9]],
                std=[x / 255.0 for x in [63.0, 62.1, 66.7]],
            ),
        ]
    )
    transform_valid = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(
                mean=[x / 255.0 for x in [125.3, 123.0, 113.9]],
                std=[x / 255.0 for x in [63.0, 62.1, 66.7]],
            ),
        ]
    )
    '''
    transform_train = transforms.Compose(
         [transforms.ToTensor(),
          transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])
    transform_valid = transforms.Compose(
         [transforms.ToTensor(),
          transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])
    
    # transform_train = transforms.Compose(
    #      [transforms.ToTensor()])
    # transform_valid = transforms.Compose(
    #      [transforms.ToTensor()])

    if args.dataset == "cifar100":

        train_dataset = datasets.CIFAR100(
            root="../data/cifar100",
            train=True,
            download=True,
            transform=transform_train,
        ) 

        mc_train_dataset = CIFAR100SubSet(
            root="../data/cifar100",
            train=True,
            download=True,
            transform=transform_train,
            returns="all",
            num_sample=args.num_mc_sample
        )


        valid_dataset = datasets.CIFAR100(
            root="../data/cifar100",
            train=False,
            download=True,
            transform=transform_valid,
        )

    else:

        train_dataset = datasets.CIFAR10(
            root="../data/cifar10",
            train=True,
            download=True,
            transform=transform_train,
        ) 

        mc_train_dataset = CIFAR10SubSet(
            root="../data/cifar10",
            train=True,
            download=True,
            transform=transform_train,
            returns="all",
            num_sample=args.num_mc_sample
        )


        valid_dataset = datasets.CIFAR10(
            root="../data/cifar10",
            train=False,
            download=True,
            transform=transform_valid,
        )



    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=bs, shuffle=True, num_workers=2
    )

    valid_loader = torch.utils.data.DataLoader(
        valid_dataset, batch_size=bs, shuffle=False, num_workers=2
    )


    def valid(head, vae, encoder, clf, data_loader, device, quit=False):
        encoder.eval()
        clf.eval()
 
        with torch.no_grad():
            correct, total = 0, 0
            for i, (inputs, labels) in enumerate(data_loader): 
                inputs, labels = inputs.to(device), labels.to(device)
                rep = head(inputs)
                (mu, std), rep_rec = vae(rep)
                outputs = clf(encoder(rep_rec))
                _, pred_label = torch.max(outputs.data, 1)

                total += inputs.data.size()[0]
                correct += (pred_label == labels.data).sum().item()
            metric = correct / float(total)
        return metric

    best_acc = 0
    for e in range(epoch_max):
        clf.train()
        head.train()
        encoder.train()
        vae.eval()
        ce_loss_list = []
        info_loss_list = []
        #pbar = tqdm(enumerate(train_loader))
        for batch_idx, (inputs, targets) in enumerate(train_loader):
            #pbar.set_description("Epoch {}".format(e+1))

            inputs, targets = inputs.to(device), targets.to(device)
            

            optim_head.zero_grad()
            optim_vae.zero_grad()
            optim_clf.zero_grad()
            optim_encoder.zero_grad()
            rep = head(inputs)

            (mu, std), rep_rec = vae(rep)

            feature = encoder(rep_rec)
            pred = clf(feature)

            # defense here

            ce_loss = criterion(pred, targets)
            # loss = -0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
            info_loss = - 0.5 * (1 + (std.pow(2)).log() - mu.pow(2) - std.pow(2)).sum(dim=1).mean()
            loss = ce_loss + args.MID_rate*info_loss


            loss.backward()

            optim_vae.step()
            optim_head.step()
            optim_clf.step()
            optim_encoder.step()
            ce_loss_list.append(ce_loss.item())
            info_loss_list.append(info_loss.item())

        clf.eval()
        head.eval()
        encoder.eval()
        vae.eval()
        train_acc = valid(head, vae, encoder, clf, train_loader, device)
        val_acc = valid(head, vae, encoder, clf, valid_loader, device)

        print(
            f"Epoch: {e}. ce_loss: {sum(ce_loss_list)/len(ce_loss_list):.4f}, info_loss: {sum(info_loss_list)/len(info_loss_list):.4f}\\"
            f"train_acc: {train_acc:.4f}, val_acc: {val_acc:.4f}"
        )
        if val_acc>best_acc:
            best_acc = val_acc
            torch.save({"head": head.state_dict(), "vae": vae.state_dict(), "encoder":encoder.state_dict(), "clf":clf.state_dict()}, args.save_models_filename)




if __name__ == "__main__":
    main()
